#!/usr/bin/env python

import pathspec
import re


class GitIgnoreToUnisonIgnore:
    """
    Class for converting a .gitignore file to unison sync ignore patterns

    The anchor path is the gitignore directory path relative to either:
        - a path, supplied via the "-path" option
        - the local root, when no "-path" is supplied
    """

    def __init__(self, anchor_path):
        self.anchor_path = anchor_path
        if not self.anchor_path.endswith("/"):
            self.anchor_path += "/"
        if self.anchor_path.startswith("/"):
            self.anchor_path = self.anchor_path[1:]

    def parse_gitignore(self, gitignore):
        spec = pathspec.PathSpec.from_lines(
            LazyCompiledGitWildMatch, gitignore.readlines()
        )
        return [
            UnisonPathIgnore.from_pattern(self.anchor_path, pattern)
            for pattern in spec.patterns
            if not pattern.is_null
        ]


class LazyCompiledGitWildMatch(pathspec.patterns.GitWildMatchPattern):
    def __init__(self, pattern, include=None):
        self.regex_s = None

        if isinstance(pattern, str):
            assert (
                include is None
            ), "include:{!r} must be null when pattern:{!r} is a string.".format(
                include, pattern
            )
            regex_s, include = self.pattern_to_regex(pattern)
            self.regex_s = regex_s
        elif pattern is not None and hasattr(pattern, "match"):
            self._regex = pattern
            self.regex_s = pattern.pattern
        elif pattern is None:
            assert (
                include is None
            ), "include:{!r} must be null when pattern:{!r} is null.".format(
                include, pattern
            )
        else:
            raise TypeError(
                "pattern:{!r} is not a string, RegexObject, or None.".format(pattern)
            )

        self.include = include

    @property
    def is_null(self):
        return self.include is None and self.regex_s is None

    @property
    def regex(self):
        if not self.is_compiled:
            self._regex = None
            if self.regex_s:
                self._regex = re.compile(self.regex_s)
        return self._regex

    @property
    def is_compiled(self):
        return hasattr(self, "_regex")


class UnisonPathIgnore:
    NON_CAPTURE_GROUP_REMOVAL_REGEX = re.compile(r"\?:")

    def __init__(self, anchor_path, regex, include):
        self.anchor_path = anchor_path
        self.raw_regex = regex
        self.include = include

    @classmethod
    def from_pattern(cls, anchor_path, pattern):
        return cls(anchor_path, pattern.regex_s, pattern.include)

    @property
    def regex(self):
        regex = self.raw_regex
        if regex.startswith("^"):
            regex = regex[1:]
        regex = re.sub(self.NON_CAPTURE_GROUP_REMOVAL_REGEX, "", regex)
        # Remove the trailing slash wildcard, unison won't match the slash
        regex = regex.replace("/.*$", "$")
        return f"^{self.anchor_path}{regex}"

    def __repr__(self):
        return (
            f"<{self.__class__.__name__}"
            f" anchor_path={self.anchor_path}"
            f" raw_regex={self.raw_regex}"
            f" include={self.include}>"
        )

    def __str__(self):
        s = "-ignore" if self.include else "-ignorenot"
        return f"{s}=Regex {self.regex}"



import logging
import os


log_level = os.environ.get("UNISON_GITIGNORE_LOG_LEVEL", "WARN")
logger = logging.getLogger("unison_gitignore")
logger.setLevel(getattr(logging, log_level))
logFormatter = logging.Formatter(fmt="%(name)s :: %(levelname)-8s :: %(message)s")
handler = logging.StreamHandler()
handler.setFormatter(logFormatter)
logger.addHandler(handler)

GIT_IGNORE = ".gitignore"


def build_cmd(cmd, unison_ignores):
    new_cmd = ["unison"] + cmd
    for unison_ignore in unison_ignores:
        new_cmd.append(str(unison_ignore))
    return new_cmd


def _is_ssh_root(root):
    return root.startswith("ssh://")


def collect_paths_from_cmd(cmd, cwd):
    if len(cmd) < 2:
        return

    if cmd[1].startswith("-"):
        # In the case of running using a profile
        logger.warning(
            "No .gitignore patterns will be added since a unison profile was given"
        )
        return

    roots = cmd[0:2]
    if not _is_ssh_root(roots[0]) and not _is_ssh_root(roots[1]):
        # Since unison matches the patterns against the path
        # and not the root, we can't add ignore rules when two local
        # paths are being synced, since they would match against both roots
        logger.warning(
            "No .gitignore patterns will be added since no remote roots were given"
        )
        return

    if _is_ssh_root(roots[0]):
        _, local_root = roots
    else:
        local_root, _ = roots

    has_path = False
    for i, token in enumerate(cmd):
        if token == "-path" and i < len(cmd):
            has_path = True
            path = cmd[i + 1]
            abs_path = os.path.join(cwd, local_root, path)
            yield abs_path, path

    if not has_path:
        abs_path = os.path.join(cwd, local_root)
        yield abs_path, ""


def collect_gitignores_from_path(path):
    for root, _, files in os.walk(path):
        if GIT_IGNORE in files:
            yield root, os.path.join(root, GIT_IGNORE)


def run_cmd(cmd):
    logger.debug(f"Running: {cmd}")
    os.execvp(cmd[0], cmd)



import pytest
from unittest import mock

from unison_gitignore.main import main


@pytest.fixture
def cmd():
    return [
        "unison_gitignore",
        "local_root",
        "ssh://remote_root",
        "-path",
        "path1",
        "-prefer",
        "root1",
    ]


@pytest.fixture
def cwd():
    return "/home/john_doe"


@pytest.fixture
def mock_collect_gitignores():
    def generator(abs_path):
        yield f"{abs_path}/stuff/data", f"{abs_path}/stuff/data/.gitignore"

    with mock.patch("unison_gitignore.main.collect_gitignores_from_path") as m:
        m.side_effect = generator
        yield m


@pytest.fixture
def mock_run_cmd():
    with mock.patch("unison_gitignore.main.run_cmd") as m:
        yield m


@pytest.fixture
def mock_files():
    with mock.patch(
        "builtins.open", mock.mock_open(read_data=mock_gitignore_contents)
    ) as mock_file:
        yield mock_file


mock_gitignore_as_array = [
    "*.py[co]",
]

mock_gitignore_contents = "\n".join(mock_gitignore_as_array)


@pytest.mark.usefixtures("mock_collect_gitignores", "mock_files")
def test_calls_run_cmd(cmd, cwd, mock_run_cmd):
    main(cmd, cwd)
    mock_run_cmd.assert_called_once()


def test_no_args_calls_run_cmd(cwd, mock_run_cmd):
    main(["unison_gitignore"], cwd)
    mock_run_cmd.assert_called_once()
    args = mock_run_cmd.call_args[0][0]
    assert args == ["unison"]


@pytest.mark.usefixtures("mock_collect_gitignores", "mock_files")
def test_calls_run_cmd_with_regex_patterns(cmd, cwd, mock_run_cmd):
    main(cmd, cwd)
    args = mock_run_cmd.call_args[0][0]
    assert args[0] == "unison"
    assert args[1:7] == cmd[1:7]
    assert len(args) == len(cmd) + len(mock_gitignore_as_array)

    for arg in args[7:]:
        assert arg.startswith("-ignore=")


@pytest.mark.usefixtures("mock_collect_gitignores", "mock_files")
@pytest.mark.parametrize(
    "cmd",
    [
        # Local ambiguity
        ["unison_gitignore", "local1", "local2", "-path", "path"],
        # Profile
        ["unison_gitignore", "my_profile", "-path", "path"],
    ],
)
def test_does_not_add_patterns_when_unable_to(cmd, cwd, mock_run_cmd):
    main(cmd, cwd)
    mock_run_cmd.assert_called_once()
    args = mock_run_cmd.call_args[0][0]
    assert args[0] == "unison"
    assert len(args) == len(cmd)
    assert args[1:] == cmd[1:]


@pytest.mark.usefixtures("mock_collect_gitignores", "mock_files")
def test_when_no_path_given_it_uses_local_root(cwd, mock_run_cmd):
    cmd = ["unison_gitignore", "/home/john_doe", "ssh://remote/john_doe_data"]
    main(cmd, cwd)
    mock_run_cmd.assert_called_once()
    args = mock_run_cmd.call_args[0][0]
    assert args[0] == "unison"
    assert args[1:3] == cmd[1:3]
    assert len(args) == 4
    assert args[3] == "-ignore=Regex ^stuff/data/(.+/)?[^/]*\\.py[co](/.*)?$"


@pytest.mark.usefixtures("mock_files")
def test_gitignore_in_root(cmd, cwd, mock_run_cmd):
    def generator(abs_path):
        yield f"{abs_path}", f"{abs_path}.gitignore"

    with mock.patch("unison_gitignore.main.collect_gitignores_from_path") as m:
        m.side_effect = generator
        main(cmd, cwd)
    mock_run_cmd.assert_called_once()
    args = mock_run_cmd.call_args[0][0]
    assert args[0] == "unison"
    assert args[1:7] == cmd[1:7]
    assert len(args) == 8
    assert args[7] == "-ignore=Regex ^(.+/)?[^/]*\\.py[co](/.*)?$"



from io import StringIO

import pytest

from unison_gitignore.parser import GitIgnoreToUnisonIgnore, LazyCompiledGitWildMatch


@pytest.fixture
def mock_path():
    return "/home/john_doe"


def test_root_path_removes_slash():
    parsed = GitIgnoreToUnisonIgnore("/").parse_gitignore(StringIO("test.py"))
    assert len(parsed) == 1
    assert parsed[0].anchor_path == ""


def test_keeps_trailing_slash_but_removes_leading_slash():
    parsed = GitIgnoreToUnisonIgnore("/a/b/c/").parse_gitignore(StringIO("test.py"))
    assert len(parsed) == 1
    assert parsed[0].anchor_path == "a/b/c/"


def test_empty_gitignore(mock_path):
    parsed = GitIgnoreToUnisonIgnore(mock_path).parse_gitignore(StringIO())
    assert parsed == []


def test_blank_lines_and_comments(mock_path):
    contents = """
    # test.py # commented out

    """
    parsed = GitIgnoreToUnisonIgnore(mock_path).parse_gitignore(StringIO(contents))
    assert parsed == []


def test_wildcard_and_negation_regex(mock_path):
    contents = """
    *.py[co]
    !test.pyc
    """
    parsed = GitIgnoreToUnisonIgnore(mock_path).parse_gitignore(StringIO(contents))
    assert len(parsed) == 2
    assert str(parsed[0]) == r"-ignore=Regex ^home/john_doe/(.+/)?[^/]*\.py[co](/.*)?$"
    assert str(parsed[1]) == r"-ignorenot=Regex ^home/john_doe/(.+/)?test\.pyc$"


def test_directories_regex(mock_path):
    contents = """
    abc/
    **/x/y
    b/**/c/d
    """
    parsed = GitIgnoreToUnisonIgnore(mock_path).parse_gitignore(StringIO(contents))
    assert len(parsed) == 3
    assert str(parsed[0]) == r"-ignore=Regex ^home/john_doe/(.+/)?abc$"
    assert str(parsed[1]) == r"-ignore=Regex ^home/john_doe/(.+/)?x/y(/.*)?$"
    assert str(parsed[2]) == r"-ignore=Regex ^home/john_doe/b(/.+)?/c/d(/.*)?$"


def test_lazy_match_does_lazy_compilation():
    match = LazyCompiledGitWildMatch(r"test\.py")
    assert not match.is_compiled
    match.regex
    assert match.is_compiled



import os
from os.path import join, exists
from setuptools import find_packages, setup

base_dir = os.path.dirname(__file__)
readme_path = join(base_dir, "README.md")
if exists(readme_path):
    with open(readme_path) as stream:
        long_description = stream.read()
else:
    long_description = ""

setup(
    name="unison-gitignore",
    install_requires=["pathspec"],
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="A unison wrapper to integrate with .gitignore",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Josh DM",
    url="https://github.com/lime-green/unison-gitignore",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    entry_points={
        "console_scripts": [
            "unison_gitignore = unison_gitignore.__main__:main",
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    license="MIT",
    keywords=["unison", "gitignore"],
)



import os
import sys

from .parser import GitIgnoreToUnisonIgnore
from .util import (
    build_cmd,
    collect_paths_from_cmd,
    collect_gitignores_from_path,
    logger,
    run_cmd,
)


def main(cmd=None, cwd=None):
    if cmd is None:
        cmd = sys.argv
    if cwd is None:
        cwd = os.getcwd()

    unison_ignores = []
    cmd_args = cmd[1:]

    for abs_path, path in collect_paths_from_cmd(cmd_args, cwd):
        for gitignore_root, gitignore in collect_gitignores_from_path(abs_path):
            gitignore_anchor = os.path.relpath(gitignore_root, abs_path)
            if gitignore_anchor == ".":
                gitignore_anchor = ""

            parser = GitIgnoreToUnisonIgnore(gitignore_anchor)
            with open(gitignore, "r") as fh:
                unison_ignores.extend(parser.parse_gitignore(fh))

    logger.info(
        f"Adding {len(unison_ignores)} ignore patterns based on .gitignore contents"
    )
    cmd_new = build_cmd(cmd_args, unison_ignores)
    run_cmd(cmd_new)
